from __future__ import annotations
from typing import Set, Dict

from PySide2 import QtWidgets, QtGui, QtCore

from sohstationviewer.database.process_db import execute_db
from sohstationviewer.view.util.one_instance_at_a_time import \
    OneWindowAtATimeDialog


def set_widget_color(widget, changed=False, read_only=False):
    """
    Set color of the given widget depend on flag read_only or changed
    :param widget: the widget to set color
    :param changed: True: red Text, white background
    :param read_only: True: grey Text, blue background (priority)
    if both flags are False: black Text, white blackground
    """
    palette = QtGui.QPalette()

    if read_only:
        # grey text
        palette.setColor(QtGui.QPalette.Text, QtGui.QColor(100, 100, 100))
        # light blue background
        palette.setColor(QtGui.QPalette.Base, QtGui.QColor(210, 240, 255))
        widget.setReadOnly(True)
        widget.setPalette(palette)
        return
    else:
        # white background
        palette.setColor(QtGui.QPalette.Base, QtGui.QColor(255, 255, 255))
    if changed:
        # red text
        palette.setColor(QtGui.QPalette.Text, QtGui.QColor(255, 0, 0))
    else:
        try:
            if widget.isReadOnly():
                # grey text
                palette.setColor(
                    QtGui.QPalette.Text, QtGui.QColor(100, 100, 100))
            else:
                # black text
                palette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
        except AttributeError:
            palette.setColor(QtGui.QPalette.Text, QtGui.QColor(0, 0, 0))
    widget.setPalette(palette)


class UiDBInfoDialog(OneWindowAtATimeDialog):
    """
    Superclass for info database dialogs under database menu.
    """

    # This line is only to type hint the class attribute inherited from
    # OneWindowAtATimeDialog.
    current_instance: UiDBInfoDialog

    def __init__(self, parent, column_headers, col_name, table_name,
                 resize_content_columns=[], required_columns={},
                 need_data_type_choice=False, check_fk=True):
        """
        :param parent: QMainWindow/QWidget - the parent widget
        :param column_headers: [str,] - headers of the columns
        :param col_name: str - key db column in the table (not count dataType
            in Channels table)
        :param table_name: str - the database table
        :param resize_content_columns: [int,] - list of indexes of columns of
            which width needs to be resize to content
        :param required_columns: {int: str,} - dict of column indexes vs column
            names of the columns that are required to have a value.
        :param need_data_type_choice: bool - to indicate if a QChoicewill be
            added
            to the top of the dialog for user to select data type.
        :param check_fk: check foreign key to turn the row to readonly if not
            satified.
        """
        self.parent = parent
        self.total_col = len(column_headers)
        self.column_headers = column_headers
        self.resize_content_columns = resize_content_columns
        self.required_columns = required_columns
        self.need_data_type_choice = need_data_type_choice
        self.col_name = col_name
        self.table_name = table_name
        self.check_fk = check_fk
        """
        changes_by_rowid: dict of changes by row ids
        """
        self.changes_by_rowid: Dict[int, Set[int]] = {}
        super(UiDBInfoDialog, self).__init__()
        # =========================== pre-define ============================
        """
        data_type: str - type of data being process
        """
        self.data_type = None

        self.setGeometry(100, 100, 900, 900)
        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)

        button_layout = self.create_buttons_section()
        main_layout.addLayout(button_layout)

        self.create_data_table_widget()
        main_layout.addWidget(self.data_table_widget, 1)
        """
        Color code is given at the end of the dialog.
        Other instructions should be given when hover over the widgets.
        """

        if self.table_name != '':
            # Not really used in this (base) class, but made available so that
            # any children can use the empty space on the bottom right of the
            # widget.
            self.bottom_layout = QtWidgets.QGridLayout()
            instruction = ("Background: LIGHT BLUE - Non Editable due to "
                           "FK constrain; "
                           "WHITE - Editable.\n"
                           "Text: BLACK - Saved; RED - Not saved.")
            # TODO: add question mark button to give instruction
            self.bottom_layout.addWidget(QtWidgets.QLabel(instruction),
                                         0, 0, 1, 1)
            main_layout.addLayout(self.bottom_layout)

    def add_widget(self, data_list, row_idx, col_idx, foreign_key=False,
                   choices=None, range_values=None, field_name='',
                   place_holder_text='') -> str:
        """
        Add a cell widget at a given row and column in the data_table_widget.

        :param data_list: list of data to be display in the data_table_widget.
            data_list is None when that column shows the row number.
        :param row_idx: index of row
        :param col_idx: index of column
        :param foreign_key: True if there is a foreign constrain that prevents
            the row to be deleted => set readonly of cell=True
        :param choices: List of choices to add to a drop down list. If
            it is available, QComboBox will be assigned for the widget.
        :param range_values: list consists of min and max values. If
            it is available, QSpinbox will be assigned for the widget
        If no choices or range available, QLineEdit  will be assigned for the
            widget.
        :param field_name: name of the column in database. If field_name=
            'convertFactor', add validator to limit precision to 6 decimal
            points
        :param place_holder_text: the text displayed in QTextEdit as a hint
            for user
        :return text: value of the current field in string
        """
        if data_list is None:
            text = str(row_idx)  # row number
        else:
            text = (str(data_list[row_idx][col_idx - 1])
                    if row_idx < len(data_list) else '')
        if data_list is None:
            widget = QtWidgets.QPushButton(text)
        elif choices is None and range_values is None:
            if field_name == 'description':
                widget = QtWidgets.QTextEdit()
                height = 12
                for line in text.split('\n'):
                    widget.append(line)
                    height += 16
                widget.setFixedHeight(height)
            else:
                widget = QtWidgets.QLineEdit(text)
                widget.setPlaceholderText(place_holder_text)
            if field_name == 'convertFactor':
                # precision=6
                validator = QtGui.QDoubleValidator(0.0, 5.0, 6)
                validator.setNotation(QtGui.QDoubleValidator.StandardNotation)
                widget.setValidator(validator)
                if text == '':
                    widget.setText('1')
        elif choices:
            widget = QtWidgets.QComboBox()
            widget.addItems(choices)
            widget.setCurrentText(text)
        elif range_values:
            widget = QtWidgets.QSpinBox()
            widget.setMinimum(range_values[0])
            widget.setMaximum(range_values[1])
            if text in ["", None, 'None']:
                widget.setValue(range_values[0])
            else:
                widget.setValue(int(text))

        if data_list is None:
            set_widget_color(widget)
            widget.setFixedWidth(40)
            widget.clicked.connect(
                lambda: self.row_number_clicked(widget))
        elif foreign_key:
            set_widget_color(widget, read_only=True)
        else:
            new = False if row_idx < len(data_list) else True
            set_widget_color(widget, read_only=False, changed=new)
            if choices is None and range_values is None:
                widget.textChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
            elif choices:
                widget.currentTextChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
            elif range_values:
                widget.valueChanged.connect(
                    lambda changed_text:
                    self.cell_input_change(changed_text, row_idx, col_idx))
        self.data_table_widget.setCellWidget(row_idx, col_idx, widget)
        if field_name == 'description':
            self.data_table_widget.resizeRowToContents(row_idx)
        return text

    @QtCore.Slot()
    def row_number_clicked(self, widget):
        """
        When a row number cell is clicked, the whole line is selected.

        :param widget: row number widget
        """
        self.data_table_widget.selectRow(int(widget.text()))
        self.data_table_widget.repaint()

    def create_buttons_section(self):
        """
        Create a horizontal layout of buttons including:
        * data_type_combo_box if self.need_data_type_choice is True
        * 'ADD ROW' button
        * 'SAVE CHANGES' button
        * 'CLOSE' button
        """
        h_layout = QtWidgets.QHBoxLayout()

        if self.need_data_type_choice:
            self.data_type_combo_box = QtWidgets.QComboBox(self)
            data_type_rows = execute_db('SELECT * FROM DataTypes')
            self.data_type_combo_box.addItems([d[0] for d in data_type_rows])
            self.data_type_combo_box.currentTextChanged.connect(
                self.data_type_changed)
            h_layout.addWidget(self.data_type_combo_box)
        if self.table_name != '':
            self.add_row_btn = QtWidgets.QPushButton(self, text='ADD ROW')
            self.add_row_btn.setFixedWidth(300)
            self.add_row_btn.clicked.connect(self.add_new_row)
            self.add_row_btn.setEnabled(False)
            h_layout.addWidget(self.add_row_btn)

            self.save_changes_btn = QtWidgets.QPushButton(
                self, text='SAVE CHANGES')
            self.save_changes_btn.setFixedWidth(300)
            self.save_changes_btn.clicked.connect(self.save_changes)
            self.save_changes_btn.setEnabled(False)
            h_layout.addWidget(self.save_changes_btn)

        self.close_btn = QtWidgets.QPushButton(self, text='CLOSE')
        self.close_btn.setFixedWidth(300)
        self.close_btn.clicked.connect(self.on_close)
        h_layout.addWidget(self.close_btn)
        return h_layout

    def on_close(self):
        if len(self.changes_by_rowid) != 0:
            msg = ("Some changes may have been created on the table.\n"
                   "Please click on Cancel to go back save those changes to "
                   "database.\n"
                   "OR click on Continue to close the dialog.")
            msgbox = QtWidgets.QMessageBox()
            msgbox.setWindowTitle('Do you want to continue?')
            msgbox.setText(msg)
            msgbox.addButton(QtWidgets.QMessageBox.Cancel)
            msgbox.addButton('Continue', QtWidgets.QMessageBox.YesRole)
            result = msgbox.exec()
            if result == QtWidgets.QMessageBox.Cancel:
                return
        self.close()

    def create_data_table_widget(self):
        """
        Create self.data_table_widget based on self.total_col, self.column
        and call update_data_table_widget_items to create cells for the widget.
        """
        self.data_table_widget = QtWidgets.QTableWidget(self)
        self.data_table_widget.verticalHeader().hide()
        self.data_table_widget.setColumnCount(self.total_col)
        self.data_table_widget.setHorizontalHeaderLabels(self.column_headers)
        header = self.data_table_widget.horizontalHeader()
        header.setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        for i in self.resize_content_columns:
            header.setSectionResizeMode(
                i, QtWidgets.QHeaderView.ResizeToContents)

        if self.need_data_type_choice:
            self.data_type = self.data_type_combo_box.currentText()
        self.update_data_table_widget_items()

    def update_data_table_widget_items(self):
        """
        Create widget cell for self.data_table_widget based on self.data_list
        for data to be displayed.
        Each row will be check for foreign_key constrain to decide readonly
        status.
        If self.data_list is empty, call clear_first_row to create a row with
        all necessary widget cells but empty.
        """
        self.data_list = self.get_data_list()
        row_count = len(self.data_list)
        row_count = 1 if row_count == 0 else row_count
        self.data_table_widget.setRowCount(row_count)
        for i in range(len(self.data_list)):
            fk = self.check_data_foreign_key(self.data_list[i][0])
            self.add_row(i, fk)
        if len(self.data_list) == 0:
            """
            No Row, should leave 1 empty row
            """
            self.clear_first_row()

        self.update()

    @QtCore.Slot()
    def add_new_row(self):
        """
        Add a new empty row to the end of the data_table_widget
        and scroll to the bottom.
        """
        row_position = self.data_table_widget.rowCount()
        self.data_table_widget.insertRow(row_position)
        self.add_row(row_position)
        self.data_table_widget.scrollToBottom()
        self.data_table_widget.repaint()  # to show row's header
        self.data_table_widget.cellWidget(row_position, 1).setFocus()

    def remove_row(self, remove_row_idx):
        """
        Remove row at remove_row_idx and move all data below it one row up.

        :param remove_row_idx: Index of row to be removed.
        """
        self.data_table_widget.removeRow(remove_row_idx)
        for i in range(remove_row_idx, self.data_table_widget.rowCount()):
            cell_widget = self.data_table_widget.cellWidget(i, 0)
            cell_widget.setText(str(i))

    @QtCore.Slot()
    def cell_input_change(self, changed_text, row_idx, col_idx):
        """
        If cell's value is changed, text's color will be red,
        otherwise, text's color will be black

        :param changed_text: new value of the cell widget
        :param row_idx: row index of the cell widget
        :param col_idx: column index of the cell widget
        """
        changed = False
        if row_idx < len(self.data_list):
            if changed_text != self.data_list[row_idx][col_idx - 1]:
                changed = True
            cell_widget = self.data_table_widget.cellWidget(row_idx, col_idx)
            set_widget_color(cell_widget, changed=changed)
        else:
            changed = True
        if changed:
            if row_idx not in self.changes_by_rowid:
                self.changes_by_rowid[row_idx] = set()
            self.changes_by_rowid[row_idx].add(col_idx - 1)  # skip order col
        else:
            try:
                self.changes_by_rowid[row_idx].remove(col_idx - 1)
            except KeyError:
                pass

    def check_data_foreign_key(self, val):
        """
        If self.check_fk is True, check foreign key constrain for self.col_name
        column in Channels data table to allow or disallow deleting record in
        Parameters or DataTypes table.

        :param val: the value that need to be checked again self.col_name
        """
        if not self.check_fk:
            return False
        sql = (f"SELECT {self.col_name} FROM channels "
               f"WHERE {self.col_name}='{val}'")
        param_rows = execute_db(sql)
        if len(param_rows) > 0:
            return True
        else:
            return False

    def reset_row_inputs(self, reset, widget_idx, list_idx):
        """
        Reset the value based on reset's value and the status of the
        cell_widget
        If there is a fk constrain, the read_only flag is True. The cell cannot
            be changed. Value should remain the same, color should remain
            grey/blue).
        If read_only flag is False the color of the widget will be set to
            black/white
        If reset=1, value at the cell widget will be reset to the original
            value in the self.data_list (database)

        :param reset: reset flag to determine how the cell widget to be reset
        :param widget_idx: index of the row in self.data_table_widget
        :param list_idx: index of the row in self.data_list
        """
        for col_idx in range(1, self.total_col):
            cell_widget = self.data_table_widget.cellWidget(
                widget_idx,
                col_idx
            )
            read_only = False
            if hasattr(cell_widget, 'isReadOnly'):
                read_only = cell_widget.isReadOnly()
            if not read_only:
                if reset == 1:
                    org_val = self.data_list[list_idx][col_idx - 1]
                    if isinstance(cell_widget, QtWidgets.QLineEdit):
                        cell_widget.setText(str(org_val))
                    elif isinstance(cell_widget, QtWidgets.QComboBox):
                        cell_widget.setCurrentText(str(org_val))
                    elif isinstance(cell_widget, QtWidgets.QSpinBox):
                        try:
                            cell_widget.setValue(int(org_val))
                        except TypeError:
                            cell_widget.setValue(0)
                set_widget_color(cell_widget)

    def add_row(self, row_idx, fk=False):
        """
        Add all widgets in a row
        """
        pass

    def data_type_changed(self):
        """
        Load new self.data_list when self.data_type_combo_box's value is
            changed
        """
        pass

    @QtCore.Slot()
    def save_changes(self):
        """
        When button 'SAVE' is clicked, check each row in data_table_widget to
        remain, update, delete, reset the row in self.data_table_widget,
        self.data_list and database.
        reset's value is the return from self.update_data. if reset=-1, the row
        remain unchanged, else self.reset_row_inputs() will be called to reset
        the row.
        """
        try:
            self.data_table_widget.focusWidget().clearFocus()
        except AttributeError:
            pass
        self.remove_count = 0
        self.insert_count = 0
        self.skip_count = 0
        row_count = self.data_table_widget.rowCount()
        for i in range(row_count):
            widget_idx = i - (row_count - self.data_table_widget.rowCount())
            if widget_idx != i and i in self.changes_by_rowid:
                # row id may not be consistent with widget_idx due to
                # row row removed or inserted.
                self.changes_by_rowid.remove(i)
                self.changes_by_rowid.add(widget_idx)
            list_idx = (i - self.remove_count
                        + self.insert_count - self.skip_count)
            try:
                row_inputs = self.get_row_inputs(widget_idx)
            except Exception as e:
                QtWidgets.QMessageBox.warning(self, "Error", str(e))
                return

            reset = self.update_data(row_inputs, widget_idx, list_idx)
            if reset > -1:
                self.reset_row_inputs(reset, widget_idx, list_idx)
            try:
                if len(self.changes_by_rowid[widget_idx]) == 0:
                    del self.changes_by_rowid[widget_idx]
            except KeyError:
                pass

    def update_data(self, row, widget_idx, list_idx, insert_sql, update_sql,
                    del_sql_add=None):
        """
        Check values of the given row to decide to add, remove, update the row
        in data_table_widget and data_list along with database
        according to the action .

        :param row: list of values of the given row in data_table_widget
        :param widget_idx: index of the rows in data_table_widget
        :param list_idx: index of the rows in self.data_list
        :param insert_sql: query to insert a row to the table
        :param update_sql: query to update a row in the table
        :param del_sql_add: additional condition to add to the common del_sql
        :return -1 for doing nothing
                0 reset color to not show value changed
                    but no need to reset input values to org row
                1 reset color to not show value changed
                    and reset input values to org row
        """
        if list_idx < len(self.data_list):
            org_row = self.data_list[list_idx]
            if row[0] == "":
                msg = (f"The {self.col_name} of '{org_row}' at "
                       f"row {widget_idx} has been changed to blank.\n\n"
                       f"Are you sure you want to delete this row in "
                       f"database?")
                result = QtWidgets.QMessageBox.question(
                    self, "Delete %s?" % org_row, msg,
                          QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
                )
                if result == QtWidgets.QMessageBox.No:
                    # reset the first widget value and call the function again
                    # to check other fields
                    cell_widget = self.data_table_widget.cellWidget(
                        widget_idx, 1
                    )
                    cell_widget.setText(org_row[0])
                    row[0] = org_row[0]
                    self.changes_by_rowid[widget_idx].remove(0)
                    self.update_data(row, widget_idx, list_idx)
                else:
                    sql = (f"DELETE FROM {self.table_name} "
                           f"WHERE {self.col_name}='{org_row[0]}'")
                    if del_sql_add is not None:
                        sql += sql
                    execute_db(sql)
                    del self.changes_by_rowid[widget_idx]
                    self.data_list.remove(org_row)
                    self.remove_row(widget_idx)
                    self.remove_count += 1
                    return -1
            if row == org_row:
                return -1
            if row[0] in [p[0] for p in self.data_list]:
                if org_row[0] != row[0]:
                    msg = (f"Row {widget_idx}: The name {self.col_name} "
                           f"has been changed to '{row[0]}' "
                           f"which already is in the database.\n\n"
                           f"It will be changed back to {org_row[0]}.")
                    QtWidgets.QMessageBox.information(self, "Error", msg)
                    # reset the first widget value and call the function again
                    # to check other fields
                    cell_widget = self.data_table_widget.cellWidget(widget_idx,
                                                                    1)
                    cell_widget.setText(org_row[0])
                    row[0] = org_row[0]
                    self.changes_by_rowid[widget_idx].remove(0)
                    self.update_data(row, widget_idx, list_idx)
                else:
                    msg = (f"Row {widget_idx}: {org_row}  has "
                           f"been changed to {row}.\n\nPlease confirm it.")
                    result = QtWidgets.QMessageBox.question(
                        self, "Confirmation", msg,
                        QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel
                    )
                    if result == QtWidgets.QMessageBox.Cancel:
                        return 1
                    else:
                        execute_db(update_sql % org_row[0])
                        self.data_list[list_idx] = row
                        del self.changes_by_rowid[widget_idx]
                        return 0

        if row[0] == "":
            msg = (f"Row {widget_idx} has blank {self.col_name}.\n\n"
                   f"It will be removed.")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            self.remove_row(widget_idx)
            del self.changes_by_rowid[widget_idx]
            return -1
        blank_required_columns = [self.required_columns[i]
                                  for i in self.required_columns.keys()
                                  if row[i - 1] == ""]
        if blank_required_columns != []:
            msg = (f"Row {widget_idx}: blank "
                   f"{', '.join(blank_required_columns)} which require some "
                   f"value(s).")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            return -1

        if row[0] in [p[0] for p in self.data_list]:
            msg = (f"Row {widget_idx}: The {self.col_name} '{row[0]}' is "
                   f"already in the database.\n\n"
                   f"Row {widget_idx} will be removed.")
            QtWidgets.QMessageBox.information(self, "Error", msg)
            self.remove_row(widget_idx)
            del self.changes_by_rowid[widget_idx]
            return -1
        execute_db(insert_sql)
        self.data_list.append(row)
        self.insert_count += 1
        del self.changes_by_rowid[widget_idx]
        return 0
